import tensorflow as tf
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.ops.partition import partition_row_indexing
from kgcnn.ops.segment import segment_ops_by_name, segment_softmax

ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdges')
class PoolingLocalEdges(GraphBaseLayer):
    r"""The main aggregation or pooling layer to collect all edges or edge-like embeddings per node,
    corresponding to the receiving node, which is defined by edge indices.
    The term pooling is here used as aggregating rather than reducing the graph as in graph pooling.

    Apply e.g. sum or mean on edges with same target ID taken from the (edge) index tensor, that has a list of
    all connections as :math:`(i, j)`. In the default definition for this layer index :math:`i` is expected ot be the
    receiving or target node (in standard case of directed edges). This can be changed by setting :obj:`pooling_index`.

    Note: index_tensor[:, :, pooling_index] is sorted for the subsequent segment-operation.
    
    Args:
        pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
        pooling_index (int): Index from edge_indices to pick ID's for pooling edge-like embeddings. Default is 0.
    """

    def __init__(self, pooling_method="mean", pooling_index=0, **kwargs):
        """Initialize layer."""
        super(PoolingLocalEdges, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.pooling_index = pooling_index

    def build(self, input_shape):
        """Build layer."""
        super(PoolingLocalEdges, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): of [node, edges, tensor_index]

                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Edge or message features of shape (batch, [M], F)
                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
    
        Returns:
            tf.RaggedTensor: Pooled feature tensor of pooled edge features for each node.
        """
        # Need ragged input but can be generalized in the future.
        self.assert_ragged_input_rank(inputs)

        nod, node_part = inputs[0].values, inputs[0].row_splits
        edge, _ = inputs[1].values, inputs[1].row_lengths()
        edgeind, edge_part = inputs[2].values, inputs[2].row_lengths()

        shiftind = partition_row_indexing(edgeind, node_part, edge_part,
                                          partition_type_target="row_splits",
                                          partition_type_index="row_length",
                                          to_indexing='batch',
                                          from_indexing=self.node_indexing)

        nodind = shiftind[:, self.pooling_index]  # Pick index eg. ingoing
        dens = edge
        if not self.is_sorted:
            node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
            nodind = tf.gather(nodind, node_order, axis=0)
            dens = tf.gather(dens, node_order, axis=0)

        # Pooling via e.g. segment_sum
        out = segment_ops_by_name(self.pooling_method, dens, nodind)

        # If not unsort_segment operation need a scatter here.
        if self.has_unconnected:
            out = tf.scatter_nd(ks.backend.expand_dims(tf.range(tf.shape(out)[0]), axis=-1), out,
                                tf.concat([tf.shape(nod)[:1], tf.shape(out)[1:]], axis=0))

        out = tf.RaggedTensor.from_row_splits(out, node_part, validate=self.ragged_validate)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingLocalEdges, self).get_config()
        config.update({"pooling_method": self.pooling_method, "pooling_index": self.pooling_index})
        return config


PoolingLocalMessages = PoolingLocalEdges  # For now, they are synonyms


@ks.utils.register_keras_serializable(package='kgcnn', name='PoolingWeightedLocalEdges')
class PoolingWeightedLocalEdges(GraphBaseLayer):
    r"""The main aggregation or pooling layer to collect all edges or edge-like embeddings per node,
    corresponding to the receiving node, which is defined by edge indices.
    The term pooling is here used as aggregating rather than reducing the graph as in graph pooling.

    Apply e.g. sum or mean on edges with same target ID taken from the (edge) index-tensor, that has a list of
    all connections as :math:`(i, j)`. In the default definition for this layer index :math:`i` is expected ot be the
    receiving or target node (in standard case of directed edges). This can be changed by setting :obj:`pooling_index`.

    .. note::
        In addition of aggregating edge embeddings a weight tensor must be supplied that scales each edge before
        pooling. Must broadcast.

    Args:
        pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
        normalize_by_weights (bool): Normalize the pooled output by the sum of weights. Default is False.
        pooling_index (int): Index to pick ID's for pooling edge-like embeddings. Default is 0.
    """

    def __init__(self, pooling_method="mean",
                 normalize_by_weights=False,
                 pooling_index=0,
                 **kwargs):
        """Initialize layer."""
        super(PoolingWeightedLocalEdges, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.normalize_by_weights = normalize_by_weights
        self.pooling_index = pooling_index

    def build(self, input_shape):
        """Build layer."""
        super(PoolingWeightedLocalEdges, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): of [node, edges, tensor_index, weights]

                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Edge or message features of shape (batch, [M], F)
                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
                - weights (tf.RaggedTensor): Edge or message weights. Must broadcast to edges or messages,
                  e.g. (batch, [M], 1)

        Returns:
            tf.RaggedTensor: Pooled feature tensor of pooled edge features for each node of shape (batch, [N], F)
        """
        self.assert_ragged_input_rank(inputs)

        nod, node_part = inputs[0].values, inputs[0].row_splits
        edge, _ = inputs[1].values, inputs[1].row_lengths()
        edgeind, edge_part = inputs[2].values, inputs[2].row_lengths()
        weights, _ = inputs[3].values, inputs[3].row_lengths()

        shiftind = partition_row_indexing(edgeind, node_part, edge_part,
                                          partition_type_target="row_splits", partition_type_index="row_length",
                                          to_indexing='batch', from_indexing=self.node_indexing)

        wval = weights
        dens = edge * wval
        nodind = shiftind[:, self.pooling_index]

        if not self.is_sorted:
            node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
            nodind = tf.gather(nodind, node_order, axis=0)
            dens = tf.gather(dens, node_order, axis=0)
            wval = tf.gather(wval, node_order, axis=0)

        # Pooling via e.g. segment_sum
        get = segment_ops_by_name(self.pooling_method, dens, nodind)

        if self.normalize_by_weights:
            get = tf.math.divide_no_nan(get, tf.math.segment_sum(wval, nodind))  # +tf.eps

        if self.has_unconnected:
            get = tf.scatter_nd(ks.backend.expand_dims(tf.range(tf.shape(get)[0]), axis=-1), get,
                                tf.concat([tf.shape(nod)[:1], tf.shape(get)[1:]], axis=0))

        out = tf.RaggedTensor.from_row_splits(get, node_part, validate=self.ragged_validate)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingWeightedLocalEdges, self).get_config()
        config.update({"pooling_method": self.pooling_method, "normalize_by_weights": self.normalize_by_weights,
                       "pooling_index": self.pooling_index})
        return config


PoolingWeightedLocalMessages = PoolingWeightedLocalEdges  # For now, they are synonyms


@ks.utils.register_keras_serializable(package='kgcnn', name='PoolingEmbedding')
class PoolingEmbedding(GraphBaseLayer):
    """Polling all embeddings of edges or nodes per batch to obtain a graph level embedding in form of a
    ::obj`tf.Tensor`.
    
    Args:
        pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
    """

    def __init__(self, pooling_method="mean", **kwargs):
        """Initialize layer."""
        super(PoolingEmbedding, self).__init__(**kwargs)
        self.pooling_method = pooling_method

    def build(self, input_shape):
        """Build layer."""
        super(PoolingEmbedding, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor): Embedding tensor of shape (batch, [N], F)
    
        Returns:
            tf.Tensor: Pooled node features of shape (batch, F)
        """
        # Need ragged input but can be generalized in the future.
        self.assert_ragged_input_rank(inputs)
        # We cast to values here
        nod, batchi = inputs.values, inputs.value_rowids()

        # Could also use reduce_sum here.
        out = segment_ops_by_name(self.pooling_method, nod, batchi)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingEmbedding, self).get_config()
        config.update({"pooling_method": self.pooling_method})
        return config


PoolingNodes = PoolingEmbedding
PoolingGlobalEdges = PoolingEmbedding


@ks.utils.register_keras_serializable(package='kgcnn', name='PoolingWeightedEmbedding')
class PoolingWeightedEmbedding(GraphBaseLayer):
    """Polling all embeddings of edges or nodes per batch to obtain a graph level embedding in form of a
    ::obj`tf.Tensor`.

    .. note::
        In addition of pooling embeddings a weight tensor must be supplied that scales each embedding before
        pooling. Must broadcast.

    Args:
        pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
    """

    def __init__(self, pooling_method="mean", **kwargs):
        """Initialize layer."""
        super(PoolingWeightedEmbedding, self).__init__(**kwargs)
        self.pooling_method = pooling_method

    def build(self, input_shape):
        """Build layer."""
        super(PoolingWeightedEmbedding, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): of [node, weights]

                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
                - weights (tf.RaggedTensor): Node or message weights. Most broadcast to nodes. Shape (batch, [N], 1).

        Returns:
            tf.Tensor: Pooled node features of shape (batch, F)
        """
        # Need ragged input but can be generalized in the future.
        self.assert_ragged_input_rank(inputs)
        # We cast to values here
        nod, batchi = inputs[0].values, inputs[0].value_rowids()
        weights, _ = inputs[1].values, inputs[1].value_rowids()
        nod = tf.math.multiply(nod, weights)
        # Could also use reduce_sum here.
        out = segment_ops_by_name(self.pooling_method, nod, batchi)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingWeightedEmbedding, self).get_config()
        config.update({"pooling_method": self.pooling_method})
        return config


PoolingWeightedNodes = PoolingWeightedEmbedding
PoolingWeightedGlobalEdges = PoolingWeightedEmbedding


@ks.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdgesLSTM')
class PoolingLocalEdgesLSTM(GraphBaseLayer):
    """The main aggregation or pooling layer to collect all edges or edge-like embeddings per node,
    corresponding to the receiving node, which is defined by edge indices.
    The term pooling is here used as aggregating rather than reducing the graph as in graph pooling.

    Here, apply LSTM on edges with same target ID taken from the (edge) index_tensor, that has a list of
    all connections as :math:`(i, j)`. In the default definition for this layer index :math:`i` is expected ot be the
    receiving or target node (in standard case of directed edges). This can be changed by setting :obj:`pooling_index`.

    Args:
        units (int): Units for LSTM cell.
        pooling_method (str): Pooling method. Default is 'LSTM', is ignored.
        pooling_index (int): Index to pick ID's for pooling edge-like embeddings. Default is 0.
        activation: Activation function to use.
            Default: hyperbolic tangent (`tanh`). If you pass `None`, no activation
            is applied (ie. "linear" activation: `a(x) = x`).
        recurrent_activation: Activation function to use for the recurrent step.
            Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
            applied (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean (default `True`), whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix, used for
            the linear transformation of the inputs. Default: `glorot_uniform`.
            recurrent_initializer: Initializer for the `recurrent_kernel` weights
            matrix, used for the linear transformation of the recurrent state.
            Default: `orthogonal`.
        bias_initializer: Initializer for the bias vector. Default: `zeros`.
            unit_forget_bias: Boolean (default `True`). If True, add 1 to the bias of
            the forget gate at initialization. Setting it to true will also force
            `bias_initializer="zeros"`. This is recommended in [Jozefowicz et
            al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf).
        kernel_regularizer: Regularizer function applied to the `kernel` weights
            matrix. Default: `None`.
        recurrent_regularizer: Regularizer function applied to the
            `recurrent_kernel` weights matrix. Default: `None`.
            bias_regularizer: Regularizer function applied to the bias vector. Default:
            `None`.
        activity_regularizer: Regularizer function applied to the output of the
            layer (its "activation"). Default: `None`.
        kernel_constraint: Constraint function applied to the `kernel` weights
            matrix. Default: `None`.
        recurrent_constraint: Constraint function applied to the `recurrent_kernel`
            weights matrix. Default: `None`.
        bias_constraint: Constraint function applied to the bias vector. Default:
            `None`.
        dropout: Float between 0 and 1. Fraction of the units to drop for the linear
            transformation of the inputs. Default: 0.
            recurrent_dropout: Float between 0 and 1. Fraction of the units to drop for
            the linear transformation of the recurrent state. Default: 0.
        return_sequences: Boolean. Whether to return the last output. in the output
            sequence, or the full sequence. Default: `False`.
        return_state: Boolean. Whether to return the last state in addition to the
            output. Default: `False`.
        go_backwards: Boolean (default `False`). If True, process the input sequence
            backwards and return the reversed sequence.
        stateful: Boolean (default `False`). If True, the last state for each sample
            at index i in a batch will be used as initial state for the sample of
            index i in the following batch.
        time_major: The shape format of the `inputs` and `outputs` tensors.
            If True, the inputs and outputs will be in shape
            `[timesteps, batch, feature]`, whereas in the False case, it will be
            `[batch, timesteps, feature]`. Using `time_major = True` is a bit more
            efficient because it avoids transposes at the beginning and end of the
            RNN calculation. However, most TensorFlow data is batch-major, so by
            default this function accepts input and emits output in batch-major
            form.
        unroll: Boolean (default `False`). If True, the network will be unrolled,
            else a symbolic loop will be used. Unrolling can speed-up a RNN, although
            it tends to be more memory-intensive. Unrolling is only suitable for short
            sequences.
    """

    def __init__(self,
                 units,
                 pooling_method="LSTM",
                 pooling_index=0,
                 activation='tanh', recurrent_activation='sigmoid',
                 use_bias=True, kernel_initializer='glorot_uniform',
                 recurrent_initializer='orthogonal',
                 bias_initializer='zeros', unit_forget_bias=True,
                 kernel_regularizer=None, recurrent_regularizer=None, bias_regularizer=None,
                 activity_regularizer=None, kernel_constraint=None, recurrent_constraint=None,
                 bias_constraint=None, dropout=0.0, recurrent_dropout=0.0,
                 return_sequences=False, return_state=False, go_backwards=False, stateful=False,
                 time_major=False, unroll=False,
                 **kwargs):
        """Initialize layer."""
        super(PoolingLocalEdgesLSTM, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.pooling_index = pooling_index

        self.lstm_unit = ks.layers.LSTM(units=units, activation=activation, recurrent_activation=recurrent_activation,
                                        use_bias=use_bias, kernel_initializer=kernel_initializer,
                                        recurrent_initializer=recurrent_initializer,
                                        bias_initializer=bias_initializer, unit_forget_bias=unit_forget_bias,
                                        kernel_regularizer=kernel_regularizer,
                                        recurrent_regularizer=recurrent_regularizer, bias_regularizer=bias_regularizer,
                                        activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint,
                                        recurrent_constraint=recurrent_constraint,
                                        bias_constraint=bias_constraint, dropout=dropout,
                                        recurrent_dropout=recurrent_dropout,
                                        return_sequences=return_sequences, return_state=return_state,
                                        go_backwards=go_backwards, stateful=stateful,
                                        time_major=time_major, unroll=unroll)
        if self.pooling_method not in ["LSTM", "lstm"]:
            print("Warning: Pooling method does not match with layer, expected 'LSTM' but got", self.pooling_method)

    def build(self, input_shape):
        """Build layer."""
        super(PoolingLocalEdgesLSTM, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [nodes, edges, tensor_index]

                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Edge or message features of shape (batch, [M], F)
                - tensor_index (tf.RaggedTensor): Edge indices of shape (batch, [M], 2)

        Returns:
            tf.RaggedTensor: Feature tensor of pooled edge features for each node of shape (batch, [N], F)
        """
        self.assert_ragged_input_rank(inputs)

        nod, node_part = inputs[0].values, inputs[0].row_splits
        edge, _ = inputs[1].values, inputs[1].row_lengths()
        edgeind, edge_part = inputs[2].values, inputs[2].row_lengths()

        shiftind = partition_row_indexing(edgeind, node_part, edge_part,
                                          partition_type_target="row_splits",
                                          partition_type_index="row_length",
                                          to_indexing='batch',
                                          from_indexing=self.node_indexing)

        nodind = shiftind[:, self.pooling_index]  # Pick first index eg. ingoing
        dens = edge
        if not self.is_sorted:
            # Sort edgeindices
            node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
            nodind = tf.gather(nodind, node_order, axis=0)
            dens = tf.gather(dens, node_order, axis=0)

        # Pooling via LSTM
        # we make a ragged input
        ragged_lstm_input = tf.RaggedTensor.from_value_rowids(dens, nodind)
        get = self.lstm_unit(ragged_lstm_input)

        if self.has_unconnected:
            # Need to fill tensor since the maximum node may not be also in pooled
            # Does not happen if all nodes are also connected
            get = tf.scatter_nd(ks.backend.expand_dims(tf.range(tf.shape(get)[0]), axis=-1), get,
                                tf.concat([tf.shape(nod)[:1], tf.shape(get)[1:]], axis=0))

        out = tf.RaggedTensor.from_row_splits(get, node_part, validate=self.ragged_validate)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingLocalEdgesLSTM, self).get_config()
        config.update({"pooling_method": self.pooling_method, "pooling_index": self.pooling_index})
        conf_lstm = self.lstm_unit.get_config()
        lstm_param = ["activation", "recurrent_activation", "use_bias", "kernel_initializer", "recurrent_initializer",
                      "bias_initializer", "unit_forget_bias", "kernel_regularizer", "recurrent_regularizer",
                      "bias_regularizer", "activity_regularizer", "kernel_constraint", "recurrent_constraint",
                      "bias_constraint", "dropout", "recurrent_dropout", "implementation", "return_sequences",
                      "return_state", "go_backwards", "stateful", "time_major", "unroll"]
        for x in lstm_param:
            config.update({x: conf_lstm[x]})
        return config


PoolingLocalMessagesLSTM = PoolingLocalEdgesLSTM  # For now they are synonyms


@ks.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdgesAttention')
class PoolingLocalEdgesAttention(GraphBaseLayer):
    r"""Pooling or aggregation of all edges or edge-like features per node,
    corresponding to node assigned by edge indices.
    Uses attention for pooling. i.e. :math:`n_i =  \sum_j \alpha_{ij} e_{ij}`
    The attention is computed via: :math:`\alpha_ij = \text{softmax}_j (a_{ij})` from the
    attention coefficients :math:`a_{ij}`.
    The attention coefficients must be computed beforehand by edge features or by :math:`\sigma( W n_i || W n_j)` and
    are passed to this layer as input. Thereby this layer has no weights and only does pooling.
    In summary, :math:`n_i = \sum_j \text{softmax}_j (a_{ij}) e_{ij}` is computed by the layer.

    An edge is defined by index tuple :math:`(i, j)` with :math:`i` being the receiving node in the default definition,
    but can be changed by pooling_index.
    Important: ID's for segment-operation and for pooling of edges are taken from edge-index-tensor.
    They are sorted for faster pooling from tensor_index[:, :, pooling_index].

    Args:
        pooling_index (int): Index to pick ID's for pooling edge-like embeddings. Default is 0.
    """

    def __init__(self, pooling_index=0, **kwargs):
        """Initialize layer."""
        super(PoolingLocalEdgesAttention, self).__init__(**kwargs)
        self.pooling_index = pooling_index

    def build(self, input_shape):
        """Build layer."""
        super(PoolingLocalEdgesAttention, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [node, edges, attention, edge_indices]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Edge or message embeddings of shape (batch, [M], F)
                - attention (tf.RaggedTensor): Attention coefficients of shape (batch, [M], 1)
                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], F)

        Returns:
            tf.RaggedTensor: Embedding tensor of pooled edge attentions for each node of shape (batch, [N], F)
        """
        # Need ragged input but can be generalized in the future.
        self.assert_ragged_input_rank(inputs)
        # We cast to values here
        nod, node_part = inputs[0].values, inputs[0].row_lengths()
        edge = inputs[1].values
        attention = inputs[2].values
        edgeind, edge_part = inputs[3].values, inputs[3].row_lengths()

        shiftind = partition_row_indexing(edgeind, node_part, edge_part, partition_type_target="row_length",
                                          partition_type_index="row_length", to_indexing='batch',
                                          from_indexing=self.node_indexing)

        nodind = shiftind[:, self.pooling_index]  # Pick first index eg. ingoing
        dens = edge
        ats = attention
        if not self.is_sorted:
            # Sort edgeindices
            node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
            nodind = tf.gather(nodind, node_order, axis=0)
            dens = tf.gather(dens, node_order, axis=0)
            ats = tf.gather(ats, node_order, axis=0)

        # Apply segmented softmax
        ats = segment_softmax(ats, nodind)
        get = dens * ats
        get = tf.math.segment_sum(get, nodind)

        if self.has_unconnected:
            # Need to fill tensor since the maximum node may not be also in pooled
            # Does not happen if all nodes are also connected
            get = tf.scatter_nd(ks.backend.expand_dims(tf.range(tf.shape(get)[0]), axis=-1), get,
                                tf.concat([tf.shape(nod)[:1], tf.shape(get)[1:]], axis=0))

        out = tf.RaggedTensor.from_row_lengths(get, node_part, validate=self.ragged_validate)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingLocalEdgesAttention, self).get_config()
        config.update({"pooling_index": self.pooling_index})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='PoolingEmbeddingAttention')
class PoolingEmbeddingAttention(GraphBaseLayer):
    r"""Polling all embeddings of edges or nodes per batch to obtain a graph level embedding in form of a
    ::obj`tf.Tensor`.

    Uses attention for pooling. i.e. :math:`s =  \sum_j \alpha_{i} n_i`.
    The attention is computed via: :math:`\alpha_i = \text{softmax}_i(a_i)` from the attention coefficients :math:`a_i`.
    The attention coefficients must be computed beforehand by edge features or by :math:`\sigma( W [s || n_i])` and
    are passed to this layer as input. Thereby this layer has no weights and only does pooling.
    In summary, :math:`s =  \sum_i \text{softmax}_j(a_i) n_i` is computed by the layer.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(PoolingEmbeddingAttention, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        super(PoolingEmbeddingAttention, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [nodes, attention]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - attention (tf.RaggedTensor): Attention coefficients of shape (batch, [N], 1)

        Returns:
            tf.Tensor: Embedding tensor of pooled node of shape (batch, F)
        """
        # Need ragged input but can be generalized in the future.
        self.assert_ragged_input_rank(inputs)
        # We cast to values here
        nod, batchi, target_len = inputs[0].values, inputs[0].value_rowids(), inputs[0].row_lengths()
        ats = inputs[1].values

        ats = segment_softmax(ats, batchi)
        get = nod * ats
        out = tf.math.segment_sum(get, batchi)

        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingEmbeddingAttention, self).get_config()
        return config


PoolingNodesAttention = PoolingEmbeddingAttention



import os
import numpy as np
import pandas as pd
from kgcnn.data.base import MemoryGraphDataset
from kgcnn.data.download import DownloadDataset


class MD17RevisedDataset(DownloadDataset, MemoryGraphDataset):
    """Store and process full MD17DatasetRevised dataset."""

    download_info = {
        "dataset_name": "MD17Revised",
        "data_directory_name": "MD17Revised",
        "download_url": "https://archive.materialscloud.org/record/file?filename=rmd17.tar.bz2&record_id=466",
        "download_file_name": 'rmd17.tar.bz2',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": "rmd17"
    }
    possible_trajectory_names = ["aspirin", "azobenzene", "benzene", "ethanol", "malonaldehyde", "naphthalene",
                                 "paracetamol", "salicylic", "toluene", "uracil"]

    def __init__(self, trajectory_name: str = None, reload=False, verbose=10):
        """Initialize MD17DatasetRevised dataset for a specific trajectory.

        Args:
            trajectory_name (str): Name of trajectory to load.
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        self.data_keys = None
        self.trajectory_name = trajectory_name
        if trajectory_name not in self.possible_trajectory_names:
            raise ValueError(
                "Name for trajectory '%s' not found. Choose: %s." % (trajectory_name, self.possible_trajectory_names))

        MemoryGraphDataset.__init__(self, dataset_name="MD17Revised", verbose=verbose)
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name,
                                           self.unpack_directory_name, "rmd17")
        self.file_name = "rmd17_%s.npz" % self.trajectory_name
        self.dataset_name = self.dataset_name + "_" + self.trajectory_name
        # May add name of trajectory to name of dataset here.

        if self.fits_in_memory:
            self.read_in_memory()

    def _get_trajectory_from_npz(self, file_path: str = None):
        if file_path is None:
            file_dir = os.path.join(self.data_directory, "npz_data")
            file_path = os.path.join(file_dir, self.file_name)
        return np.load(file_path)

    def _get_train_test_splits(self):
        file_dir = os.path.join(self.data_directory, "splits")

        def read_splits(file_name: str) -> list:
            return [
                np.squeeze(pd.read_csv(
                    os.path.join(file_dir, file_name % i), header=None).values, axis=-1) for i in range(1, 6)]
        return read_splits("index_train_0%i.csv"), read_splits("index_test_0%i.csv")

    def read_in_memory(self,):
        """Read dataset trajectory into memory.

        Returns:
            self.
        """
        data = self._get_trajectory_from_npz()
        self.data_keys = list(data.keys())
        for key in ["coords", "energies", "forces", "old_indices", "old_energies", "old_forces"]:
            self.assign_property(key, [x for x in data[key]])
        node_number = data["nuclear_charges"]
        self.assign_property("nuclear_charges", [np.array(node_number) for _ in range(len(self))])

        # Add splits to self.
        splits_train, splits_test = self._get_train_test_splits()
        property_train = []
        property_test = []
        for i in range(len(self)):
            is_train = []
            is_test = []
            for j, split in enumerate(splits_train):
                if i in split:
                    is_train.append(j + 1)
            for j, split in enumerate(splits_test):
                if i in split:
                    is_test.append(j + 1)
            # Add to list
            if len(is_train) > 0:
                property_train.append(np.array(is_train, dtype="int"))
            else:
                property_train.append(None)
            if len(is_test) > 0:
                property_test.append(np.array(is_test, dtype="int"))
            else:
                property_test.append(None)
        self.assign_property("train", property_train)
        self.assign_property("test", property_test)

        return self



import os.path
import numpy as np
from sklearn.preprocessing import StandardScaler as StandardScalerSklearn
from kgcnn.data.utils import save_json_file, load_json_file


class StandardScaler(StandardScalerSklearn):
    r"""Standard scaler that inherits from :obj:`sklearn.preprocessing.StandardScaler`.
    Added functionality to save and load weights of this scaler. Included dummy kwarg 'atomic_number' to be compatible
    with molar scaler.

    .. code-block:: python

        import numpy as np
        from kgcnn.scaler.scaler import StandardScaler
        data = np.random.rand(5).reshape((5,1))
        scaler = StandardScaler()
        scaler.fit(X=data)
        print(scaler.get_weights())
        print(scaler.get_config())
        print(scaler.inverse_transform(scaler.transform(X=data)))
        print(data)
        scaler.save("example.json")
        new_scaler = StandardScaler()
        new_scaler.load("example.json")
        print(new_scaler.inverse_transform(scaler.transform(X=data)))
    """

    _attributes_list_sklearn = ["n_features_in_", "mean_", "scale_", "var_", "feature_names_in_", "n_samples_seen_"]

    def __init__(self, *, copy=True, with_mean=True, with_std=True):
        super(StandardScaler, self).__init__(copy=copy, with_mean=with_mean, with_std=with_std)

    def fit(self, X, *, y=None, sample_weight=None, atomic_number=None):
        return super(StandardScaler, self).fit(X=X, y=y, sample_weight=sample_weight)

    def partial_fit(self, X, y=None, sample_weight=None, atomic_number=None):
        # For partial fit internally uses args and not kwargs.
        # Can not request kwargs after argument X here.
        return super(StandardScaler, self).partial_fit(X=X, y=y, sample_weight=sample_weight)

    def fit_transform(self, X, *, y=None, atomic_number=None, **fit_params):
        return super(StandardScaler, self).fit_transform(X=X, y=y, **fit_params)

    def transform(self, X, *, copy=None, atomic_number=None):
        return super(StandardScaler, self).transform(X=X, copy=copy)

    def inverse_transform(self, X, *, copy=None, atomic_number=None):
        return super(StandardScaler, self).inverse_transform(X=X, copy=copy)

    def get_scaling(self):
        """Get scale of shape (1, n_properties)."""
        if not hasattr(self, "scale_"):
            return
        scale = np.array(self.scale_)
        scale = np.expand_dims(scale, axis=0)
        return scale

    def get_config(self):
        """Get configuration for scaler."""
        config = super(StandardScaler, self).get_params()
        return config

    def set_config(self, config):
        """Set configuration for scaler.

        Args:
            config (dict): Config dictionary.
        """
        self.set_params(**config)

    def get_weights(self) -> dict:
        """Get weights for this scaler after fit."""
        weight_dict = dict()
        for x in self._attributes_list_sklearn:
            if hasattr(self, x):
                value = getattr(self, x)
                value_update = {x: np.array(value).tolist()} if value is not None else {x: value}
                weight_dict.update(value_update)
        return weight_dict

    def set_weights(self, weights: dict):
        """Set weights for this scaler.

        Args:
            weights (dict): Weight dictionary.
        """
        for item, value in weights.items():
            if item in self._attributes_list_sklearn:
                setattr(self, item, np.array(value))
            else:
                print("`StandardScaler` got unknown weight '%s'." % item)

    def save_weights(self, file_path: str):
        """Save weights as numpy to file.

        Args:
            file_path: Filepath to save weights.
        """
        weights = self.get_weights()
        # Make them all numpy arrays for save.
        for key, value in weights.items():
            weights[key] = np.array(value)
        if len(weights) > 0:
            np.savez(os.path.splitext(file_path)[0] + ".npz", **weights)
        else:
            print("Error no weights to save for `StandardScaler`.")

    def save(self, file_path: str):
        """Save scaler serialization to file.

        Args:
            file_path: Filepath to save scaler serialization.
        """
        conf = self.get_config()
        weights = self.get_weights()
        full_info = {"class_name": type(self).__name__, "module_name": type(self).__module__,
                     "config": conf, "weights": weights}
        save_json_file(full_info, os.path.splitext(file_path)[0] + ".json")

    def load(self, file_path: str):
        """Load scaler serialization from file.

        Args:
            file_path: Filepath to load scaler serialization.
        """
        full_info = load_json_file(file_path)
        # Could verify class_name and module_name here.
        self.set_config(full_info["config"])
        self.set_weights(full_info["weights"])


import numpy as np
import tensorflow as tf
import math
import logging
ks = tf.keras

logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class LinearWarmUpScheduler(ks.callbacks.LearningRateScheduler):
    r"""Warmup scheduler that increases the learning rate over the first steps or epochs.
    Inherits from :obj:`ks.callbacks.LearningRateScheduler`. Note that the parameter :obj:`schedule` in the
    constructor can be `None` which sets the wam-up in :obj:`linear_warmup_schedule_epoch_lr` only.
    Another learning rate schedule can be provided that makes use of :obj:`linear_warmup_schedule_epoch_lr` in
    subclasses.

    .. note::

        To increase the learning rate within each epoch, you must provide :obj:`steps_per_epoch`.
    """

    def __init__(self, schedule=None, verbose: int = 0, steps_per_epoch: int = None,
                 lr_start: float = None, epo_warmup: int = 0):
        r"""Initialize class.

        Args:
            schedule (Callable): Learning rate schedule. Default is None.
            verbose (int): Verbosity. Default is 0.
            steps_per_epoch (int): Steps per epoch. Required for sub-epoch increase. Default is None.
            lr_start (float): Learning rate to reach in linear ramp to start learning. Default is None.
            epo_warmup (int): Number of warmup epochs. Default is None.
        """
        # This is for passing other schedule through warmup to `ks.callbacks.LearningRateScheduler`.
        # IDEA: We could pass a restart option.
        if schedule is None:
            schedule = self.linear_warmup_schedule_epoch_lr
        self.verbose = verbose
        super(LinearWarmUpScheduler, self).__init__(
            schedule=schedule, verbose=verbose)
        self.epo_warmup = max(epo_warmup, 0)
        self.__warming_up = False
        self.steps_per_epoch = steps_per_epoch
        self.lr_start = lr_start
        if self.steps_per_epoch is None:
            module_logger.warning("`steps_per_epoch` is not set. Can't increase lr during epochs of warmup.")

    def linear_warmup_schedule_epoch_lr(self, epoch, lr):
        """Learning rate schedule function for warmup.

        Returns the current learning rate unchanged apart from warmup period.

        Args:
            epoch (int): Epoch index (integer, indexed from 0).
            lr (float): Current learning rate.

        Returns:
            float: New learning rate.
        """
        if epoch < self.epo_warmup:
            self.__warming_up = True
            new_lr = float(self.lr_start * epoch / self.epo_warmup)
        else:
            self.__warming_up = False
            new_lr = lr
        return new_lr

    def on_train_batch_begin(self, batch, logs=None):
        """Recursive increase of the learning rate warmup between epochs.

        Args:
            batch (int): Batch index (integer, indexed from 0).
            logs: Not used.

        Returns:
            None
        """
        if self.__warming_up and self.steps_per_epoch is not None:
            if not hasattr(self.model.optimizer, "lr"):
                raise ValueError('Optimizer must have a "lr" attribute.')
            lr = float(ks.backend.get_value(self.model.optimizer.lr))
            lr = lr + self.lr_start / self.epo_warmup / self.steps_per_epoch
            if batch > self.steps_per_epoch:
                module_logger.warning("Found `batch` > `steps_per_epoch` during warmup.")
            if self.verbose > 0:
                print("{0}/{1}: Warmup-step increase lr to {2}.".format(batch, self.steps_per_epoch, lr))
            ks.backend.set_value(self.model.optimizer.lr, ks.backend.get_value(lr))

    def get_config(self):
        """Get config for this class."""
        config = super(LinearWarmUpScheduler, self).get_config()
        config.update({"lr_start": self.lr_start, "epo_warmup": self.epo_warmup,
                       "verbose": self.verbose, "steps_per_epoch": self.steps_per_epoch})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='CosineAnnealingLRScheduler')
class CosineAnnealingLRScheduler(ks.callbacks.LearningRateScheduler):
    r"""Callback for cosine learning rate (LR) schedule. This class inherits from
    :obj:`ks.callbacks.LearningRateScheduler` and applies :obj:`schedule_epoch_lr`.
    Proposed by `SGDR <https://arxiv.org/abs/1608.03983>`_.
    The cosine part without restarts for the LR Schedule follows:

    .. math::

        \eta_t (T_{cur}) = \eta_{min} + \frac{1}{2}(\eta_{max} - \eta_{min})\left(1 +
        \cos\left(\frac{T_{cur}}{T_{max}}\pi\right)\right)
    """

    def __init__(self, lr_start: float, epoch_max: int, lr_min: float = 0, verbose: int = 0):
        """Set the parameters for the learning rate scheduler.

        Args:
            lr_start (float): Learning rate at the start of the annealing. decay.
            epoch_max (int): Maximum number of iterations.
            lr_min (float): Minimum learning rate allowed during the decay. Default is 0.0.
            verbose (int): Verbosity. Default is 0.
        """
        self.epoch_max = max(epoch_max, 0)
        self.lr_min = lr_min
        self.lr_start = lr_start
        self.verbose = verbose
        super(CosineAnnealingLRScheduler, self).__init__(
            schedule=self.schedule_epoch_lr, verbose=verbose)

    def schedule_epoch_lr(self, epoch, lr):
        """Closed from of learning rate."""
        new_lr = self.lr_min + (self.lr_start - self.lr_min) * (
                1 + math.cos(math.pi * epoch / self.epoch_max)) / 2
        return float(new_lr)

    def get_config(self):
        """Get config for this class."""
        config = super(CosineAnnealingLRScheduler, self).get_config()
        config.update({"lr_start": self.lr_start, "epoch_max": self.epoch_max,
                       "lr_min": self.lr_min, "verbose": self.verbose})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='LinearWarmupExponentialLRScheduler')
class LinearWarmupExponentialLRScheduler(LinearWarmUpScheduler):
    r"""Callback for exponential learning rate schedule with warmup. This class inherits from
    :obj:`LinearWarmUpScheduler`, which inherits from :obj:`ks.callbacks.LearningRateScheduler`.
    The learning rate :math:`\eta` is reduced or increased (usually :math:`\gamma < 1`) after
    warmup epochs :math:`T_0` as:

    .. math::

        \eta (T) = \eta_0 \; \gamma^{T-T_0}
    """

    def __init__(self, lr_start: float, gamma: float, epo_warmup: int = 10, lr_min: float = 0.0,
                 verbose: int = 0, steps_per_epoch: int = None):
        """Set the parameters for the learning rate scheduler.

        Args:
            lr_start (float): Learning rate at the start of the exponential decay.
            gamma (float): Multiplicative factor of learning rate decay.
            epo_warmup (int): Number of warmup epochs. Default is 10.
            lr_min (float): Minimum learning rate allowed during the decay and start. Default is 0.0.
            verbose (int): Verbosity. Default is 0.
            steps_per_epoch (int): Number of steps per epoch. Required for warmup to linearly increase between epochs.
                Default is None.
        """
        self.gamma = gamma
        self.lr_min = lr_min
        super(LinearWarmupExponentialLRScheduler, self).__init__(
            schedule=self.schedule_epoch_lr, verbose=verbose, lr_start=lr_start, epo_warmup=epo_warmup,
            steps_per_epoch=steps_per_epoch)

    def schedule_epoch_lr(self, epoch, lr):
        """Change the learning rate after warmup exponentially.

        Args:
            epoch (int): Epoch index (integer, indexed from 0).
            lr (float): Current learning rate. Not used.

        Returns:
            float: New learning rate.
        """
        new_lr = self.lr_start * (self.gamma ** (epoch - self.epo_warmup))
        new_lr = self.linear_warmup_schedule_epoch_lr(epoch, new_lr)
        return max(float(new_lr), self.lr_min)

    def get_config(self):
        """Get config for this class."""
        config = super(LinearWarmupExponentialLRScheduler, self).get_config()
        config.update({"gamma": self.gamma, "lr_min": self.lr_min})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='LinearWarmupExponentialLearningRateScheduler')
class LinearWarmupExponentialLearningRateScheduler(LinearWarmUpScheduler):
    r"""Callback for exponential learning rate schedule with warmup. This class inherits from
    :obj:`LinearWarmUpScheduler`, which inherits from :obj:`ks.callbacks.LearningRateScheduler`.
    The learning rate :math:`\eta` is reduced after warmup epochs :math:`T_0` with lifetime :math:`\tau` as:

    .. math::

        \eta (T) = \eta_0 \; e^{-\frac{T-T_0}{\tau}}
    """

    def __init__(self, lr_start: float, decay_lifetime: float, epo_warmup: int = 10, lr_min: float = 0.0,
                 verbose: int = 0, steps_per_epoch: int = None):
        """Set the parameters for the learning rate scheduler.

        Args:
            lr_start (float): Learning rate at the start of the exp. decay.
            decay_lifetime (float): Tau or lifetime parameter in the exponential in epochs.
            epo_warmup (int): Number of warmup epochs. Default is 10.
            lr_min (float): Minimum learning rate allowed during the decay. Default is 0.0.
            verbose (int): Verbosity. Default is 0.
            steps_per_epoch (int): Number of steps per epoch. Required for warmup to linearly increase between epochs.
                Default is None.
        """
        self.decay_lifetime = decay_lifetime
        self.lr_min = lr_min
        super(LinearWarmupExponentialLearningRateScheduler, self).__init__(
            schedule=self.schedule_epoch_lr, verbose=verbose, lr_start=lr_start, epo_warmup=epo_warmup,
            steps_per_epoch=steps_per_epoch)

    def schedule_epoch_lr(self, epoch, lr):
        """Reduce the learning rate after warmup exponentially.

        Args:
            epoch (int): Epoch index (integer, indexed from 0).
            lr (float): Current learning rate. Not used.

        Returns:
            float: New learning rate.
        """
        new_lr = float(self.lr_start * np.exp(-(epoch - self.epo_warmup) / self.decay_lifetime))
        new_lr = self.linear_warmup_schedule_epoch_lr(epoch, new_lr)
        return max(new_lr, self.lr_min)

    def get_config(self):
        """Get config for this class."""
        config = super(LinearWarmupExponentialLearningRateScheduler, self).get_config()
        config.update({"decay_lifetime": self.decay_lifetime, "lr_min": self.lr_min})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='LinearLearningRateScheduler')
class LinearLearningRateScheduler(ks.callbacks.LearningRateScheduler):
    r"""Callback for linear change of the learning rate. This class inherits from
    :obj:`ks.callbacks.LearningRateScheduler`.
    The learning rate :math:`\eta_0` is reduced linearly at :math:`T_0` epochs up to :math:`T_{max}`
    epochs to reach the learning rate :math:`\eta_{T_{max}}`:

    .. math::

        \eta (T) = \eta_0 - \frac{\eta_0 - \eta_{T_{max}}}{T_{max}-T_0} (T-T_0)  \;\; \text{for} \;\; T>T_0
    """

    def __init__(self, learning_rate_start: float = 1e-3, learning_rate_stop: float = 1e-5, epo_min: int = 0,
                 epo: int = 500, verbose: int = 0, eps: float = 1e-8):
        """Set the parameters for the learning rate scheduler.

        Args:
            learning_rate_start (float): Initial learning rate. Default is 1e-3.
            learning_rate_stop (float): End learning rate. Default is 1e-5.
            epo_min (int): Minimum number of epochs to keep the learning-rate constant before decrease. Default is 0.
            epo (int): Total number of epochs. Default is 500.
            eps (float): Minimum learning rate. Default is 1e-08.
            verbose (int): Verbosity. Default is 0.
        """
        super(LinearLearningRateScheduler, self).__init__(schedule=self.schedule_epoch_lr, verbose=verbose)
        self.learning_rate_start = learning_rate_start
        self.learning_rate_stop = learning_rate_stop
        self.epo = epo
        self.epo_min = epo_min
        self.eps = float(eps)

    def schedule_epoch_lr(self, epoch, lr):
        r"""Reduce the learning linearly.

        Kept constant for :obj:`epo_min` before decrease.

        Args:
            epoch (int): Epoch index (integer, indexed from 0).
            lr (float): Current learning rate. Not used.

        Returns:
            float: New learning rate.
        """
        if epoch < self.epo_min:
            out = float(self.learning_rate_start)
        else:
            out = float(self.learning_rate_start - (self.learning_rate_start - self.learning_rate_stop) / (
                    self.epo - self.epo_min) * (epoch - self.epo_min))
        return max(float(out), self.eps)

    def get_config(self):
        """Get config for this class."""
        config = super(LinearLearningRateScheduler, self).get_config()
        config.update({"learning_rate_start": self.learning_rate_start,
                       "learning_rate_stop": self.learning_rate_stop,
                       "epo": self.epo, "epo_min": self.epo_min, "eps": self.eps})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='LinearWarmupLinearLearningRateScheduler')
class LinearWarmupLinearLearningRateScheduler(LinearWarmUpScheduler):
    r"""Callback for linear change of the learning rate with warmup. This class inherits from
    :obj:`LinearWarmUpScheduler`, which inherits from :obj:`ks.callbacks.LearningRateScheduler`.
    The learning rate is increased in a warmup phase of :math:`T_0` epochs up to :math:`\eta_0`.
    The learning rate :math:`\eta_0` is reduced linearly at :math:`T_0` epochs up to :math:`T_{max}`
    epochs to reach the learning rate :math:`\eta_{T_{max}}`:

    .. math::

        \eta (T) = \eta_0 - \frac{\eta_0 - \eta_{T_{max}}}{T_{max}-T_0} (T-T_0)  \;\; \text{for} \;\; T>T_0
    """

    def __init__(self, learning_rate_start: float = 1e-3, learning_rate_stop: float = 1e-5, epo_warmup: int = 0,
                 epo: int = 500, verbose: int = 0, eps: float = 1e-8, steps_per_epoch: int = None,
                 lr_start: int = None):
        """Set the parameters for the learning rate scheduler.

        Args:
            learning_rate_start (float): Initial learning rate. Default is 1e-3.
            learning_rate_stop (float): End learning rate. Default is 1e-5.
            epo (int): Total number of epochs. Default is 500.
            eps (float): Minimum learning rate. Default is 1e-08.
            verbose (int): Verbosity. Default is 0.
            steps_per_epoch (int): Number of steps per epoch. Required for warmup to linearly increase between epochs.
                Default is None.
            lr_start (int): Ignored, set to `learning_rate_start`.
        """
        super(LinearWarmupLinearLearningRateScheduler, self).__init__(
            schedule=self.schedule_epoch_lr, verbose=verbose, lr_start=learning_rate_start, epo_warmup=epo_warmup,
            steps_per_epoch=steps_per_epoch)
        self.learning_rate_start = learning_rate_start
        self.learning_rate_stop = learning_rate_stop
        self.epo = epo
        self.eps = float(eps)

    def schedule_epoch_lr(self, epoch, lr):
        """Reduce the learning linearly after warmup.

        Args:
            epoch (int): Epoch index (integer, indexed from 0).
            lr (float): Current learning rate. Not used.

        Returns:
            float: New learning rate.
        """
        new_lr = self.learning_rate_start - (self.learning_rate_start - self.learning_rate_stop) / (
                self.epo - self.epo_warmup) * (epoch - self.epo_warmup)
        new_lr = self.linear_warmup_schedule_epoch_lr(epoch, new_lr)
        return max(float(new_lr), self.eps)

    def get_config(self):
        """Get config for this class."""
        config = super(LinearWarmupLinearLearningRateScheduler, self).get_config()
        config.update({
            "learning_rate_start": self.learning_rate_start, "learning_rate_stop": self.learning_rate_stop,
            "epo": self.epo, "eps": self.eps})
        return config



import tensorflow as tf
from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.geom import NodeDistanceEuclidean, EdgeAngle, BesselBasisLayer, NodePosition, ShiftPeriodicLattice
from kgcnn.layers.modules import DenseEmbedding, OptionalInputEmbedding, LazyConcatenate, LazySubtract, LazyAdd
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.utils.models import update_model_kwargs
from kgcnn.layers.conv.dimenet_conv import SphericalBasisLayer, EmbeddingDimeBlock
from kgcnn.layers.conv.mxmnet_conv import MXMGlobalMP, MXMLocalMP

ks = tf.keras

# Implementation of MXMNet in `tf.keras` from paper:
# Molecular Mechanics-Driven Graph Neural Network with Multiplex Graph for Molecular Structures
# by Shuo Zhang, Yang Liu, Lei Xie (2020)
# https://arxiv.org/abs/2011.07457
# https://github.com/zetayue/MXMNet


model_default = {
    "name": "MXMNet",
    "inputs": [{"shape": (None, ), "name": "node_number", "dtype": "float32", "ragged": True},
               {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32", "ragged": True},
               {"shape": (None, ), "name": "edge_attributes", "dtype": "float32", "ragged": True},
               {"shape": (None, 2), "name": "edge_indices", "dtype": "int64", "ragged": True},
               {"shape": [None, 2], "name": "angle_indices_1", "dtype": "int64", "ragged": True},
               {"shape": [None, 2], "name": "angle_indices_2", "dtype": "int64", "ragged": True},
               {"shape": (None, 2), "name": "range_indices", "dtype": "int64", "ragged": True}],
    "input_embedding": {"node": {"input_dim": 95, "output_dim": 32,
                                 "embeddings_initializer": {
                                     "class_name": "RandomUniform",
                                     "config": {"minval": -1.7320508075688772, "maxval": 1.7320508075688772}}},
                        "edge": {"input_dim": 32, "output_dim": 32}},
    "bessel_basis_local": {"num_radial": 16, "cutoff": 5.0, "envelope_exponent": 5},
    "bessel_basis_global": {"num_radial": 16, "cutoff": 5.0, "envelope_exponent": 5},  # Should match range_indices
    "spherical_basis_local": {"num_spherical": 7, "num_radial": 6, "cutoff": 5.0, "envelope_exponent": 5},
    "mlp_rbf_kwargs": {"units": 32, "activation": "swish"},
    "mlp_sbf_kwargs": {"units": 32, "activation": "swish"},
    "global_mp_kwargs": {"units": 32},
    "local_mp_kwargs": {"units": 32, "output_units": 1, "output_kernel_initializer": "zeros"},
    "use_edge_attributes": False,
    "depth": 3,
    "verbose": 10,
    "node_pooling_args": {"pooling_method": "sum"},
    "output_embedding": "graph", "output_to_tensor": True,
    "use_output_mlp": True,
    "output_mlp": {"use_bias": [True], "units": [1],
                   "activation": ["linear"]}
}


@update_model_kwargs(model_default)
def make_model(inputs: list = None,
               input_embedding: dict = None,
               depth: int = None,
               name: str = None,
               bessel_basis_local: dict = None,
               bessel_basis_global: dict = None,
               spherical_basis_local: dict = None,
               use_edge_attributes: bool = None,
               mlp_rbf_kwargs: dict = None,
               mlp_sbf_kwargs: dict = None,
               global_mp_kwargs: dict = None,
               local_mp_kwargs: dict = None,
               verbose: int = None,
               output_embedding: str = None,
               use_output_mlp: bool = None,
               node_pooling_args: dict = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None
               ):
    r"""Make `MXMNet <https://arxiv.org/abs/2011.07457>`_ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.MXMNet.model_default`.

    Inputs:
        list: `[node_attributes, node_coordinates, edge_attributes, edge_indices, angle_indices_1, angle_indices_2,
         range_indices]`

            - node_attributes (tf.RaggedTensor): Node attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_attributes (tf.RaggedTensor): Edge attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_indices (tf.RaggedTensor): Index list for (local) edges of shape `(batch, None, 2)`.
            - range_indices (tf.RaggedTensor): Index list for (global) range-like edges of shape `(batch, None, 2)`.
            - node_coordinates (tf.RaggedTensor): Node (atomic) coordinates of shape `(batch, None, 3)`.
            - angle_indices_1 (tf.RaggedTensor): Index list of angles referring to (local) edge connections of
              shape `(batch, None, 2)`. Angles of edge pairing (ij, jk).
            - angle_indices_2 (tf.RaggedTensor): Index list of angles referring to (local) edge connections of
              shape `(batch, None, 2)`.  Angles of edge pairing (ij, ik).

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in :obj:`Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        bessel_basis_local: Dictionary of layer arguments unpacked in local `:obj:BesselBasisLayer` layer.
        bessel_basis_global: Dictionary of layer arguments unpacked in global `:obj:BesselBasisLayer` layer.
        spherical_basis_local: Dictionary of layer arguments unpacked in `:obj:SphericalBasisLayer` layer.
        use_edge_attributes: Whether to add edge attributes. Default is False.
        mlp_rbf_kwargs: Dictionary of layer arguments unpacked in `:obj:MLP` layer for RBF feed-forward.
        mlp_sbf_kwargs: Dictionary of layer arguments unpacked in `:obj:MLP` layer for SBF feed-forward.
        global_mp_kwargs: Dictionary of layer arguments unpacked in `:obj:MXMGlobalMP` layer.
        local_mp_kwargs: Dictionary of layer arguments unpacked in `:obj:MXMLocalMP` layer.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    edge_input = ks.layers.Input(**inputs[2])
    edge_index_input = ks.layers.Input(**inputs[3])
    angle_index_input_1 = ks.layers.Input(**inputs[4])
    angle_index_input_2 = ks.layers.Input(**inputs[5])
    range_index_input = ks.layers.Input(**inputs[6])

    # Rename to short names and make embedding, if no feature dimension.
    x = xyz_input
    n = EmbeddingDimeBlock(**input_embedding["node"])(node_input) if len(inputs[0]["shape"]) < 2 else node_input
    ed = OptionalInputEmbedding(**input_embedding['edge'], use_embedding=len(inputs[2]['shape']) < 2)(edge_input)
    ei_l = edge_index_input
    ri_g = range_index_input
    ai_1 = angle_index_input_1
    ai_2 = angle_index_input_2

    # Calculate distances and spherical and bessel basis for local edges including angles.
    # For the first version, we restrict ourselves to 2-hop angles.
    pos1_l, pos2_l = NodePosition()([x, ei_l])
    d_l = NodeDistanceEuclidean()([pos1_l, pos2_l])
    rbf_l = BesselBasisLayer(**bessel_basis_local)(d_l)
    v12_l = LazySubtract()([pos1_l, pos2_l])
    a_l_1 = EdgeAngle()([v12_l, ai_1])
    a_l_2 = EdgeAngle(vector_scale=[1.0, -1.0])([v12_l, ai_2])
    sbf_l_1 = SphericalBasisLayer(**spherical_basis_local)([d_l, a_l_1, ai_1])
    sbf_l_2 = SphericalBasisLayer(**spherical_basis_local)([d_l, a_l_2, ai_2])

    # Calculate distance and bessel basis for global (range) edges.
    pos1_g, pos2_g = NodePosition()([x, ri_g])
    d_g = NodeDistanceEuclidean()([pos1_g, pos2_g])
    rbf_g = BesselBasisLayer(**bessel_basis_global)(d_g)

    if use_edge_attributes:
        rbf_l = LazyConcatenate()([rbf_l, ed])

    rbf_l = GraphMLP(**mlp_rbf_kwargs)(rbf_l)
    sbf_l_1 = GraphMLP(**mlp_sbf_kwargs)(sbf_l_1)
    sbf_l_2 = GraphMLP(**mlp_sbf_kwargs)(sbf_l_2)
    rbf_g = GraphMLP(**mlp_rbf_kwargs)(rbf_g)

    # Model
    h = n
    nodes_list = []
    for i in range(0, depth):
        h = MXMGlobalMP(**global_mp_kwargs)([h, rbf_g, ri_g])
        h, t = MXMLocalMP(**local_mp_kwargs)([h, rbf_l, sbf_l_1, sbf_l_2, ei_l, ai_1, ai_2])
        nodes_list.append(t)

    # Output embedding choice
    out = LazyAdd()(nodes_list)
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)(out)
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)(out)
        if output_to_tensor:  # For tf version < 2.8 cast to tensor below.
            out = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported output embedding for mode `MXMNet`")

    model = ks.models.Model(
        inputs=[node_input, xyz_input, edge_input, edge_index_input, angle_index_input_1, angle_index_input_2,
                range_index_input],
        outputs=out,
        name=name
    )
    return model



import os
import pickle
import numpy as np
import scipy.io
import json
import pandas as pd
from typing import Union
from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset


class QM8Dataset(QMDataset, DownloadDataset):
    """Store and process QM8 dataset."""

    download_info = {
        "dataset_name": "QM8",
        "data_directory_name": "qm8",
        "download_url": "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb8.tar.gz",
        "download_file_name": 'gdb8.tar.gz',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": "gdb8"
    }

    def __init__(self, reload: bool = False, verbose: int = 1):
        """Initialize QM8 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 0=silent. Default is 1.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM8")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = [
            "E1-CC2", "E2-CC2", "f1-CC2", "f2-CC2", "E1-PBE0", "E2-PBE0","f1-PBE0",
            "f2-PBE0","E1-PBE0", "E2-PBE0", "f1-PBE0", "f2-PBE0", "E1-CAM", "E2-CAM", "f1-CAM", "f2-CAM"
        ]
        self.label_units = ["[?]"]*len(self.label_names)
        self.label_unit_conversion = np.array([1.0] * 14)  # Pick always same units for training
        self.dataset_name = "QM8"
        self.require_prepare_data = False
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name, self.unpack_directory_name)
        self.file_name = "qm8.csv"

        if not os.path.exists(self.file_path):
            original_name = os.path.join(self.data_directory, "qm8.sdf.csv")
            if os.path.exists(original_name):
                os.rename(original_name, self.file_path)

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory(label_column_name=self.label_names)


# ds = QM8Dataset(reload=True)


import tensorflow as tf
import kgcnn.ops.activ
from kgcnn.layers.modules import DenseEmbedding, ActivationEmbedding, DropoutEmbedding
from kgcnn.layers.norm import GraphBatchNormalization, GraphLayerNormalization
from kgcnn.layers.base import GraphBaseLayer

ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='MLPBase')
class MLPBase(GraphBaseLayer):
    r"""Base class for multilayer perceptron that consist of multiple feed-forward networks.

    This base class simply manages layer arguments for :obj:`MLP`. They contain arguments for :obj:`Dense`,
    :obj:`Dropout` and :obj:`BatchNormalization` or :obj:`LayerNormalization`,
    since MLP is made up of stacked :obj:`Dense` layers with optional normalization and
    dropout to improve stability or regularization. Here, a list in place of arguments must be provided that applies
    to each layer. If not a list is given, then the single argument is used for each layer.
    The number of layers is determined by :obj:`units` argument, which should be list.

    Hence, this base class holds arguments for batch-normalization which should be applied between kernel
    and activation. And dropout after the kernel output and before normalization.

    This base class does not initialize any sub-layers or implements :obj:`call`, only for managing arguments.

    """

    def __init__(self,
                 units,
                 use_bias=True,
                 activation=None,
                 activity_regularizer=None,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_constraint=None,
                 bias_constraint=None,
                 # Normalization
                 use_normalization=False,
                 normalization_technique="batch",
                 axis=-1,
                 momentum=0.99,
                 epsilon=0.001,
                 center=True,
                 scale=True,
                 beta_initializer='zeros',
                 gamma_initializer='ones',
                 moving_mean_initializer='zeros',
                 moving_variance_initializer='ones',
                 beta_regularizer=None,
                 gamma_regularizer=None,
                 beta_constraint=None,
                 gamma_constraint=None,
                 # Dropout
                 use_dropout=False,
                 rate=None,
                 noise_shape=None,
                 seed=None,
                 **kwargs):
        r"""Initialize with parameter for MLP layer that match :obj:`Dense` layer, including :obj:`Dropout` and
        :obj:`BatchNormalization` or :obj:`LayerNormalization`.

        Args:
            units: Positive integer, dimensionality of the output space.
            activation: Activation function to use.
                If you don't specify anything, no activation is applied
                (ie. "linear" activation: `a(x) = x`).
            use_bias: Boolean, whether the layer uses a bias vector.
            kernel_initializer: Initializer for the `kernel` weights matrix.
            bias_initializer: Initializer for the bias vector.
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix.
            bias_regularizer: Regularizer function applied to the bias vector.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation").
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix.
            bias_constraint: Constraint function applied to the bias vector.
            use_normalization: Whether to use a normalization layer in between.
            normalization_technique: Which keras normalization technique to apply.
                This can be either 'batch', 'layer', 'group' etc.
            axis: Integer, the axis that should be normalized (typically the features
                axis). For instance, after a `Conv2D` layer with
                `data_format="channels_first"`, set `axis=1` in `GraphBatchNormalization`.
            momentum: Momentum for the moving average.
            epsilon: Small float added to variance to avoid dividing by zero.
            center: If True, add offset of `beta` to normalized tensor. If False, `beta`
                is ignored.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used. When the
                next layer is linear (also e.g. `nn.relu`), this can be disabled since the
                scaling will be done by the next layer.
            beta_initializer: Initializer for the beta weight.
            gamma_initializer: Initializer for the gamma weight.
            moving_mean_initializer: Initializer for the moving mean.
            moving_variance_initializer: Initializer for the moving variance.
            beta_regularizer: Optional regularizer for the beta weight.
            gamma_regularizer: Optional regularizer for the gamma weight.
            beta_constraint: Optional constraint for the beta weight.
            gamma_constraint: Optional constraint for the gamma weight.
            use_dropout: Whether to use dropout layers in between.
            rate: Float between 0 and 1. Fraction of the input units to drop.
            noise_shape: 1D integer tensor representing the shape of the
                binary dropout mask that will be multiplied with the input.
                For instance, if your inputs have shape`(batch_size, timesteps, features)` and
                you want the dropout mask to be the same for all timesteps,
                you can use `noise_shape=(batch_size, 1, features)`.
            seed: A Python integer to use as random seed.
        """
        super(MLPBase, self).__init__(**kwargs)
        local_kw = locals()

        # List for groups of arguments.
        key_list_act = ["activation", "activity_regularizer"]
        key_list_dense = ["units", "use_bias", "kernel_regularizer", "bias_regularizer",
                          "kernel_initializer", "bias_initializer", "kernel_constraint", "bias_constraint"]
        key_list_norm = ["axis", "momentum", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer",
                         "moving_mean_initializer", "moving_variance_initializer", "beta_regularizer",
                         "gamma_regularizer",
                         "beta_constraint", "gamma_constraint"]
        key_list_dropout = ["rate", "noise_shape", "seed"]
        key_list_use = ["use_dropout", "use_normalization", "normalization_technique"]
        self._key_list_init = ["kernel_initializer", "bias_initializer", "beta_initializer", "gamma_initializer",
                               "moving_mean_initializer", "moving_variance_initializer"]
        self._key_list_reg = ["activity_regularizer", "kernel_regularizer", "bias_regularizer",
                              "beta_regularizer", "gamma_regularizer"]
        self._key_list_const = ["kernel_constraint", "bias_constraint", "beta_constraint", "gamma_constraint"]

        # Dictionary of kwargs for MLP.
        self._key_list = key_list_act + key_list_dense + key_list_norm + key_list_dropout + key_list_use
        mlp_kwargs = {key: local_kw[key] for key in self._key_list}

        # Everything should be defined by units.
        if isinstance(units, int):
            units = [units]
        if not isinstance(units, list):
            raise ValueError("Units must be a list or a single int for `MLP`.")
        self._depth = len(units)
        # Special case, if axis is supposed to be multiple axis, use tuple here.
        if not isinstance(axis, list):
            axis = [axis for _ in units]
        # Special case, for shape, use tuple here.
        if not isinstance(noise_shape, list):
            noise_shape = [noise_shape for _ in units]

        # Assert matching number of args
        def assert_args_is_list(args):
            if not isinstance(args, (list, tuple)):
                return [args for _ in range(self._depth)]
            return args

        # Make every argument to list.
        for key, value in mlp_kwargs.items():
            mlp_kwargs[key] = assert_args_is_list(value)

        # Check correct length for all arguments.
        for key, value in mlp_kwargs.items():
            if len(units) != len(value):
                raise ValueError("Provide matching list of units %s and %s or simply a single value." % (units, key))

        # Deserialize initializer, regularizes, constraints and activation.
        for sl, sm in [
            (self._key_list_init, ks.initializers.get), (self._key_list_reg, ks.regularizers.get),
            (self._key_list_const, ks.constraints.get), (["activation"], ks.activations.get)
        ]:
            for key in sl:
                mlp_kwargs[key] = [sm(x) for x in mlp_kwargs[key]]

        # Assign to self as '_conf_'.
        for key, value in mlp_kwargs.items():
            setattr(self, "_conf_" + key, list(value))

        # Processed arguments for each layer.
        self._conf_mlp_dense_layer_kwargs = [{"units": self._conf_units[i],
                                              "use_bias": self._conf_use_bias[i],
                                              "name": self.name + "_dense_" + str(i),
                                              "activation": "linear",
                                              "activity_regularizer": None,
                                              "kernel_regularizer": self._conf_kernel_regularizer[i],
                                              "bias_regularizer": self._conf_bias_regularizer[i],
                                              "kernel_initializer": self._conf_kernel_initializer[i],
                                              "bias_initializer": self._conf_bias_initializer[i],
                                              "kernel_constraint": self._conf_kernel_constraint[i],
                                              "bias_constraint": self._conf_bias_constraint[i]}
                                             for i in range(self._depth)]
        self._conf_mlp_activ_layer_kwargs = [{"activation": self._conf_activation[i],
                                              "name": self.name + "_act_" + str(i),
                                              "activity_regularizer": self._conf_activity_regularizer[i]}
                                             for i in range(self._depth)]
        self._conf_mlp_norm_layer_kwargs = [{"name": self.name + '_norm_' + str(i),
                                             "axis": self._conf_axis[i],
                                             "epsilon": self._conf_epsilon[i],
                                             "center": self._conf_center[i],
                                             "scale": self._conf_scale[i],
                                             "beta_initializer": self._conf_beta_initializer[i],
                                             "gamma_initializer": self._conf_gamma_initializer[i],
                                             "beta_regularizer": self._conf_beta_regularizer[i],
                                             "gamma_regularizer": self._conf_gamma_regularizer[i],
                                             "beta_constraint": self._conf_beta_constraint[i],
                                             "gamma_constraint": self._conf_gamma_constraint[i]}
                                            for i in range(self._depth)]
        self._conf_mlp_batch_layer_kwargs = [{"name": self.name + '_norm_' + str(i),
                                             "axis": self._conf_axis[i],
                                             "epsilon": self._conf_epsilon[i],
                                             "center": self._conf_center[i],
                                             "scale": self._conf_scale[i],
                                             "beta_initializer": self._conf_beta_initializer[i],
                                             "gamma_initializer": self._conf_gamma_initializer[i],
                                             "beta_regularizer": self._conf_beta_regularizer[i],
                                             "gamma_regularizer": self._conf_gamma_regularizer[i],
                                             "beta_constraint": self._conf_beta_constraint[i],
                                             "gamma_constraint": self._conf_gamma_constraint[i],
                                             "momentum": self._conf_momentum[i],
                                             "moving_mean_initializer": self._conf_moving_mean_initializer[i],
                                             "moving_variance_initializer": self._conf_moving_variance_initializer[i]
                                              }
                                            for i in range(self._depth)]
        self._conf_mlp_drop_layer_kwargs = [{"name": self.name + '_drop_' + str(i),
                                             "rate": self._conf_rate[i],
                                             "noise_shape": self._conf_noise_shape[i],
                                             "seed": self._conf_seed[i]}
                                            for i in range(self._depth)]

    def build(self, input_shape):
        """Build layer."""
        super(MLPBase, self).build(input_shape)

    def get_config(self):
        """Update config."""
        config = super(MLPBase, self).get_config()
        for key in self._key_list:
            config.update({key: getattr(self, "_conf_" + key)})

        # Serialize initializer, regularizes, constraints and activation.
        for sl, sm in [
            (self._key_list_init, ks.initializers.serialize), (self._key_list_reg, ks.regularizers.serialize),
            (self._key_list_const, ks.constraints.serialize), (["activation"], ks.activations.serialize)
        ]:
            for key in sl:
                config.update({key: [sm(x) for x in getattr(self, "_conf_" + key)]})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='MLP')
class MLP(MLPBase):
    r"""Multilayer perceptron that consist of multiple :obj:`Dense` layers.

    .. note::
        Please see layer arguments of :obj:`MLPBase` for configuration!

    This layer adds normalization and dropout for normal tensor input. Please, see keras
    `documentation <https://www.tensorflow.org/api_docs/python/tf>`_ of
    :obj:`Dense`, :obj:`Dropout`, :obj:`BatchNormalization` and :obj:`LayerNormalization` for more information.

    Additionally, graph oriented normalization is supported. You can choose :obj:`normalization_technique` to be
    either 'BatchNormalization', 'LayerNormalization', 'GraphLayerNormalization', or 'GraphBatchNormalization'.

    """

    def __init__(self, units, **kwargs):
        """Initialize MLP. See MLPBase."""
        super(MLP, self).__init__(units=units, **kwargs)

        self.mlp_dense_layer_list = [DenseEmbedding(
            **self._conf_mlp_dense_layer_kwargs[i]) for i in range(self._depth)]

        self.mlp_activation_layer_list = [ActivationEmbedding(
            **self._conf_mlp_activ_layer_kwargs[i]) for i in range(self._depth)]

        self.mlp_dropout_layer_list = [
            DropoutEmbedding(**self._conf_mlp_drop_layer_kwargs[i]) if self._conf_use_dropout[i] else None for i
            in range(self._depth)]

        self.mlp_norm_layer_list = [None] * self._depth
        for i in range(self._depth):
            if self._conf_use_normalization[i]:
                if self._conf_normalization_technique[i] in ["batch", "BatchNormalization"]:
                    self.mlp_norm_layer_list[i] = ks.layers.BatchNormalization(
                        **self._conf_mlp_batch_layer_kwargs[i])
                elif self._conf_normalization_technique[i] in ["graph_batch", "GraphBatchNormalization"]:
                    self.mlp_norm_layer_list[i] = GraphBatchNormalization(
                        **self._conf_mlp_batch_layer_kwargs[i])
                elif self._conf_normalization_technique[i] in ["layer", "LayerNormalization"]:
                    self.mlp_norm_layer_list[i] = ks.layers.LayerNormalization(
                        **self._conf_mlp_norm_layer_kwargs[i])
                elif self._conf_normalization_technique[i] in ["graph_layer", "GraphLayerNormalization"]:
                    self.mlp_norm_layer_list[i] = GraphLayerNormalization(
                        **self._conf_mlp_norm_layer_kwargs[i])
                else:
                    raise NotImplementedError(
                        "Normalization via %s not supported." % self._conf_normalization_technique[i])

    def build(self, input_shape):
        """Build layer."""
        super(MLP, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (tf.Tensor): Input tensor with last dimension not `None`.

        Returns:
            tf.Tensor: MLP forward pass.
        """
        x = inputs
        for i in range(len(self._conf_units)):
            x = self.mlp_dense_layer_list[i](x, **kwargs)
            if self._conf_use_dropout[i]:
                x = self.mlp_dropout_layer_list[i](x, **kwargs)
            if self._conf_use_normalization[i]:
                x = self.mlp_norm_layer_list[i](x, **kwargs)
            x = self.mlp_activation_layer_list[i](x, **kwargs)
        out = x
        return out

    def get_config(self):
        """Update config."""
        config = super(MLP, self).get_config()
        return config


GraphMLP = MLP



import unittest
import numpy as np

from kgcnn.graph.geom import range_neighbour_lattice
# from kgcnn.data.datasets.MatProjectEFormDataset import MatProjectEFormDataset


class TestRangePeriodic(unittest.TestCase):
    # Handmade test case.
    artificial_lattice = np.array([[1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    artificial_atoms = np.array([[0.1, 0.0, 0.0], [0.5, 0.5, 0.5]])

    # Example from dataset, where lattice and atom coordinates do not really align.
    real_lattice = np.array([[-8.71172704, -0., -5.02971843],
                             [-10.97279872, -0.01635133, 8.94600922],
                             [-6.5538005, 12.48246168, 1.29207947]])
    real_atoms = np.array([[-24.14652308, 12.46611035, 6.41607351],
                           [-2.09180318, 0., -1.20770325],
                           [0., 0., 0.],
                           [-4.35586352, 0., -2.51485921]])

    def test_nn_range(self):

        def _test(atom, lattice, max_r):
            indices_max, images_max, dist_max = range_neighbour_lattice(atom,
                                                                        lattice, max_distance=max_r,
                                                                        max_neighbours=None)
            # For simplicity pick atom 0 to check NN.
            dist_max_0 = dist_max[indices_max[:, 0] == 0]

            test_x_results = []
            for x in [5, 10, 50, 100, 500, 1000]:
                if len(dist_max_0) < x:
                    print("Too small R for NN. Please increase for test.")
                indices, images, dist = range_neighbour_lattice(atom, lattice, max_distance=None, max_neighbours=x)
                test_num = len(indices) == len(atom) * x
                dist_0 = dist[indices[:, 0] == 0]
                test_correct_nn = np.amax(np.abs(dist_max_0[: len(dist_0)] - dist_0)) < 1e-6
                if not test_correct_nn:
                    print(np.abs(dist_max_0[: len(dist_0)] - dist_0))
                test_x_results.append(test_num and test_correct_nn)
            return test_x_results

        self.assertTrue(all(_test(self.artificial_atoms, self.artificial_lattice, 5.0)))
        self.assertTrue(all(_test(self.real_atoms, self.real_lattice, 100.0)))

    def test_dist_range(self):

        test_distance = [3.0, 4.0, 10.0, 20.0]

        def _test(atoms, lattice, ref_num_dist):
            # ref_num_dist = [len(self.compare_reference(atoms, lattice, x)[0])
            #                 for x in test_distance]
            num_dist = [len(range_neighbour_lattice(atoms,
                                                    lattice, max_distance=x,
                                                    max_neighbours=None)[0])
                        for x in test_distance]
            test_results = []
            for x, y in zip(num_dist, ref_num_dist):
                # print(x, y)
                test_results.append(x == y)
            return test_results

        self.assertTrue(all(_test(self.artificial_atoms, self.artificial_lattice, [484, 1072, 16736, 133816])))
        self.assertTrue(all(_test(self.real_atoms, self.real_lattice, [8, 8, 24, 320])))

    # def test_dist_all_correct(self):
    #     indices, images, dist = range_neighbour_lattice(self.artificial_atoms,
    #                                                     self.artificial_lattice, max_distance=5.0,
    #                                                     max_neighbours=None)
    #     indices, images, dist = self.full_sort(indices, images, dist)
    #
    #     ref_indices, ref_images, ref_dist = self.compare_reference(self.artificial_atoms,
    #     self.artificial_lattice, 5.0)
    #     ref_indices, ref_images, ref_dist = self.full_sort(ref_indices, ref_images, ref_dist)
    #     self.assertTrue(np.amax(np.abs(dist - ref_dist)) < 1e-6)
    #     self.assertTrue(np.amax(np.abs(ref_images - images)) < 1e-6)
    #     self.assertTrue(np.amax(np.abs(ref_indices - indices)) < 1e-6)
    #
    #     indices, images, dist = range_neighbour_lattice(self.real_atoms,
    #                                                     self.real_lattice, max_distance=5.0,
    #                                                     max_neighbours=None)
    #     if len(indices) <= 0:
    #         return
    #     indices, images, dist = self.full_sort(indices, images, dist)
    #
    #     ref_indices, ref_images, ref_dist = self.compare_reference(self.real_atoms, self.real_lattice, 5.0)
    #     ref_indices, ref_images, ref_dist = self.full_sort(ref_indices, ref_images, ref_dist)
    #     self.assertTrue(np.amax(np.abs(dist - ref_dist)) < 1e-6)
    #     self.assertTrue(np.amax(np.abs(ref_images - images)) < 1e-6)
    #     self.assertTrue(np.amax(np.abs(ref_indices - indices)) < 1e-6)

    def set_real_lattice_from_data(self, data, i):
        self.real_lattice = data[i]["graph_lattice"]
        self.real_atoms = data[i]["node_coordinates"]

    @staticmethod
    def full_sort(indices, images, dist):
        def reorder(order, *args):
            return [x[order] for x in args]

        s = np.argsort(dist, kind="stable")
        indices, images, dist = reorder(s, indices, images, dist)
        s = np.argsort(images[:, 2], kind="stable")
        indices, images, dist = reorder(s, indices, images, dist)
        s = np.argsort(images[:, 1], kind="stable")
        indices, images, dist = reorder(s, indices, images, dist)
        s = np.argsort(images[:, 0], kind="stable")
        indices, images, dist = reorder(s, indices, images, dist)
        s = np.argsort(indices[:, 1], kind="stable")
        indices, images, dist = reorder(s, indices, images, dist)
        s = np.argsort(indices[:, 0], kind="stable")
        indices, images, dist = reorder(s, indices, images, dist)
        return indices, images, dist

    # @staticmethod
    # def compare_reference(coordinates, lattice, max_distance):
    #     from pymatgen.core.structure import Structure
    #     py_struct = Structure(lattice, species=["C"] * len(coordinates),
    #                           coords=np.dot(coordinates, np.linalg.inv(lattice)))
    #     all_nbrs = py_struct.get_all_neighbors_py(max_distance, include_index=True)
    #
    #     all_edge_distance = []
    #     all_edge_indices = []
    #     all_edge_image = []
    #     for i, start_site in enumerate(all_nbrs):
    #         edge_distance = []
    #         edge_indices = []
    #         edge_image = []
    #         for j, stop_site in enumerate(start_site):
    #             edge_distance.append(stop_site.nn_distance)
    #             edge_indices.append([i, stop_site.index])
    #             edge_image.append(stop_site.image)
    #         # Sort after distance
    #         edge_distance = np.array(edge_distance)
    #         order_dist = np.argsort(edge_distance)
    #         edge_distance = edge_distance[order_dist]
    #         edge_indices = np.array(edge_indices, dtype="int")[order_dist]
    #         edge_image = np.array(edge_image, dtype="int")[order_dist]
    #         # Append to index list
    #         if len(edge_distance) > 0:
    #             all_edge_distance.append(edge_distance)
    #             all_edge_indices.append(edge_indices)
    #             all_edge_image.append(edge_image)
    #
    #     if len(all_edge_distance) > 0:
    #         all_edge_distance = np.concatenate(all_edge_distance, axis=0)
    #     if len(all_edge_indices) > 0:
    #         all_edge_indices = np.concatenate(all_edge_indices, axis=0)
    #     if len(all_edge_image) > 0:
    #         all_edge_image = np.concatenate(all_edge_image, axis=0)
    #     return all_edge_indices, all_edge_image, all_edge_distance


# dataset = MatProjectEFormDataset()

if __name__ == '__main__':
    # test = TestRangePeriodic()
    # test.test_dist_range()
    # test.test_nn_range()
    # for i in range(20000, len(dataset)):
    #     if i % 1000 == 0:
    #         print("..%s" % i)
    #     test.set_real_lattice_from_data(dataset, i)
    #     test.test_nn_range()
    unittest.main()



import tensorflow as tf
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.gather import GatherNodesOutgoing, GatherState, GatherEmbeddingSelection
from kgcnn.layers.pooling import PoolingLocalEdgesAttention, PoolingNodes, PoolingNodesAttention
from kgcnn.layers.modules import LazySubtract, DenseEmbedding, DropoutEmbedding, LazyConcatenate, ActivationEmbedding
from kgcnn.layers.conv.mpnn_conv import GRUUpdate
# import tensorflow.keras as ks
# import tensorflow.python.keras as ks
ks = tf.keras

# Gated recurrent unit update. See kgcnn.layers.conv.mpnn_conv for details.
HamNetGRUUnion = GRUUpdate


@ks.utils.register_keras_serializable(package='kgcnn', name='HamNetNaiveUnion')
class HamNetNaiveUnion(GraphBaseLayer):
    r"""Simple union that concatenates a feature tensor :math:`\mathbf{x}` and its updates :math:`\mathbf{x}_u`
    and applies a fully connected dense layer,
    i.e. a linear transformation with weights :math:`\mathbf{W}^{\top}`, :math:`\mathbf{b}` plus activation
    :math:`\sigma`.

    .. math::
        \mathbf{x}^{\prime} = \sigma \left[ \left( \mathbf{x} \; || \; \mathbf{x}_u \right) \mathbf{W}^{\top} +
        \mathbf{b} \right]

    """
    def __init__(self,
                 units: int,
                 activation="kgcnn>leaky_relu",
                 use_bias: bool = True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        r"""Initialize layer with arguments of :obj:`ks.layers.Dense`.

        Args:
            units (int): Positive integer, dimensionality of the output space.
            activation: Activation function to use.
                If you don't specify anything, no activation is applied
                (ie. "linear" activation: `a(x) = x`). Default is "kgcnn>leaky_relu".
            use_bias (bool): Boolean, whether the layer uses a bias vector. Default is True.
            kernel_initializer: Initializer for the `kernel` weights matrix. Default is "glorot_uniform".
            bias_initializer: Initializer for the bias vector. Default is "zeros".
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix. Default is None.
            bias_regularizer: Regularizer function applied to the bias vector. Default is None.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation"). Default is None.
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix. Default is None.
            bias_constraint: Constraint function applied to the bias vector. Default is None.
        """
        super(HamNetNaiveUnion, self).__init__(**kwargs)
        self.units = int(units)
        self.use_bias = use_bias
        kernel_args = {"kernel_regularizer": kernel_regularizer,
                       "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
        self.lay_dense = DenseEmbedding(units=units, activation=activation, use_bias=use_bias, **kernel_args)
        self.lay_concat = LazyConcatenate()

    def build(self, input_shape):
        """Build layer."""
        super(HamNetNaiveUnion, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [nodes, node_updates]

                - nodes (tf.RaggedTensor): Node features of shape `(batch, [N], F)`
                - node_updates (tf.RaggedTensor): Node features of shape `(batch, [N], F)`

        Returns:
            tf.RaggedTensor: Embedding tensor of updated node features of shape `(batch, [N], F)`.
        """
        n, nu = inputs
        nnu = self.lay_concat([n, nu], **kwargs)
        n_out = self.lay_dense(nnu, **kwargs)
        return n_out

    def get_config(self):
        """Update layer config."""
        config = super(HamNetNaiveUnion, self).get_config()
        config.update({"units": self.units, "use_bias": self.use_bias})
        conf_dense = self.lay_dense.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "use_bias", "activation"]:
            config.update({x: conf_dense[x]})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='HamNetGlobalReadoutAttend')
class HamNetGlobalReadoutAttend(GraphBaseLayer):
    r"""Computes attentive updates for fingerprint generation according to `HamNet <https://arxiv.org/abs/2105.03688>`_.
    The naming convention follows the authors `implementation <https://github.com/PKUterran/MoleculeClub>`_.
    The layer is used in :obj:`HamNetFingerprintGenerator` and computes the attentive state updates.
    The node features are first transformed by a :obj:`Dense` layer:
    :math:`\mathbf{h}' = \sigma\;(\mathbf{h} \mathbf{W}^T)` which yields the attention coefficients from state
    :math:`\mathbf{s}`:

    .. math::
        a_i = w^T [\mathbf{h}_i' \; || \; \mathbf{s}]

    with :math:`\alpha_i = \text{softmax}({a_i \; | \; i \in V})` the final state update :math:`\mathbf{m}`:

    .. math::
        \mathbf{m} = \sigma \; \sum_i \alpha_i \mathbf{h}'_i

    Update :math:`\mathbf{m}` is returned by the layer. Here, :math:`\sigma` denotes an activation function.

    """

    def __init__(self,
                 units,
                 activation="kgcnn>leaky_relu",
                 activation_last="elu",
                 use_bias=True,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 use_dropout=False,
                 rate=None, noise_shape=None, seed=None,
                 **kwargs):
        """Initialize layer.

        Args:
            units (int): Units for the linear transformation of node features before attention.
            activation (str, dict): Activation. Default is "kgcnn>leaky_relu".
            activation_last (str, dict): Last activation for messages. Default is "elu".
            use_bias (bool): Boolean, whether the layer uses a bias vector. Default is True.
            kernel_initializer: Initializer for the `kernel` weights matrix. Default is "glorot_uniform".
            bias_initializer: Initializer for the bias vector. Default is "zeros".
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix. Default is None.
            bias_regularizer: Regularizer function applied to the bias vector. Default is None.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation"). Default is None.
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix. Default is None.
            bias_constraint: Constraint function applied to the bias vector. Default is None.
            use_dropout (bool): Whether to use dropout on input features. Default is False.
            rate (float): Float between 0 and 1. Fraction of the input units to drop.
            noise_shape: 1D integer tensor representing the shape of the
                binary dropout mask that will be multiplied with the input.
            seed (int): A Python integer to use as random seed.
        """
        super(HamNetGlobalReadoutAttend, self).__init__(**kwargs)
        self.units = int(units)
        self.use_bias = use_bias
        self.use_dropout = use_dropout
        kernel_args = {"kernel_regularizer": kernel_regularizer,
                       "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
        self.gather_state = GatherState()
        if self.use_dropout:
            self.dropout_layer = DropoutEmbedding(rate=rate, noise_shape=noise_shape, seed=seed)
        self.dense_attend = DenseEmbedding(units=units, activation=activation, use_bias=use_bias, **kernel_args)
        self.dense_align = DenseEmbedding(1, activation="linear", use_bias=use_bias, **kernel_args)
        self.lay_concat = LazyConcatenate(axis=-1)
        self.pool_attention = PoolingNodesAttention()
        self.final_activ = ActivationEmbedding(activation=activation_last,
                                               activity_regularizer=activity_regularizer)

    def build(self, input_shape):
        """Build layer."""
        super(HamNetGlobalReadoutAttend, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [state, nodes]

                - state (tf.Tensor): Molecular embedding of shape `(batch, F)`
                - nodes (tf.RaggedTensor): Node features of shape `(batch, [N], F)`

        Returns:
            tf.RaggedTensor: Embedding tensor of pooled node attentions of shape (batch, F)
        """
        hm_ftr, hv_ftr = inputs
        hm_v_ftr = self.gather_state([hm_ftr, hv_ftr], **kwargs)

        attend_ftr = hv_ftr
        if self.use_dropout:
            attend_ftr = self.dropout_layer(attend_ftr, **kwargs)
        attend_ftr = self.dense_attend(attend_ftr, **kwargs)
        align_ftr = self.lay_concat([hm_v_ftr, hv_ftr], **kwargs)
        if self.use_dropout:
            align_ftr = self.dropout_layer(align_ftr, **kwargs)
        align_ftr = self.dense_align(align_ftr, **kwargs)
        mm_ftr = self.pool_attention([attend_ftr, align_ftr], **kwargs)
        mm_ftr = self.final_activ(mm_ftr, **kwargs)
        return mm_ftr, align_ftr

    def get_config(self):
        """Update layer config."""
        config = super(HamNetGlobalReadoutAttend, self).get_config()
        config.update({"use_bias": self.use_bias, "units": self.units, "use_dropout": self.use_dropout})
        conf_sub = self.dense_attend.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
            config.update({x: conf_sub[x]})
        if self.use_dropout:
            conf_drop = self.dropout_layer.get_config()
            for x in ["rate", "noise_shape", "seed"]:
                config.update({x: conf_drop[x]})
        conf_last = self.final_activ.get_config()
        config.update({"activation_last": conf_last["activation"]})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='HamNetFingerprintGenerator')
class HamNetFingerprintGenerator(GraphBaseLayer):
    r"""Computes readout or fingerprint generation according to `HamNet <https://arxiv.org/abs/2105.03688>`_.
    The naming convention follows the authors `implementation <https://github.com/PKUterran/MoleculeClub>`_.
    The layer generates a molecular or global message by iteratively updating from node embeddings. Initial state
    :math:`\mathbf{s}^0 = \frac{1}{n} \sum_i \sigma (\mathbf{h} W^T)` is updated :math:`l=1\dots L` times from
    messages :math:`\mathbf{m}^l` via a gated recurrent unit and subsequent activation :math:`\sigma`:

    .. math::
        \mathbf{s}^{l+1} = \sigma \left[\; \text{GRU}(\mathbf{s}^l, \mathbf{m}^{l}) \;\right]

    The message is obtained from an attentive readout function :math:`f` which is implemented here in
    :obj:`HamNetGlobalReadoutAttend`:

    .. math::
        \mathbf{m}^{l+1} = f(\mathbf{h}, \mathbf{m}^l)

    The final embedding :math:`\mathbf{s}^L` is used as output or molecular state.

    """

    def __init__(self,
                 units: int,
                 units_attend: int,
                 activation="kgcnn>leaky_relu",
                 use_bias: bool = True,
                 depth=4,
                 pooling_method="mean",
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 recurrent_activation='sigmoid',
                 recurrent_initializer='orthogonal',
                 recurrent_regularizer=None,
                 recurrent_constraint=None,
                 dropout=0.0,
                 recurrent_dropout=0.0,
                 reset_after=True,
                 use_dropout=False,
                 rate=None, noise_shape=None, seed=None,
                 **kwargs):
        """Initialize layer.

        Args:
            units (int): Units for the linear transformation of node features before attention.
            units_attend (int): Units for attention attributes.
            activation (str, dict): Activation. Default is "kgcnn>leaky_relu".
            use_bias (bool): Boolean, whether the layer uses a bias vector. Default is True.
            depth (int): Number of iterations. Default is 4.
            pooling_method(str): Initial pooling before iteration. Default is "mean".
            kernel_initializer: Initializer for the `kernel` weights matrix. Default is "glorot_uniform".
            bias_initializer: Initializer for the bias vector. Default is "zeros".
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix. Default is None.
            bias_regularizer: Regularizer function applied to the bias vector. Default is None.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation"). Default is None.
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix. Default is None.
            bias_constraint: Constraint function applied to the bias vector. Default is None.
            recurrent_activation: Activation function to use for the recurrent step.
                Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
                applied (ie. "linear" activation: `a(x) = x`).
            recurrent_initializer: Initializer for the `recurrent_kernel`
                weights matrix, used for the linear transformation of the recurrent state.
                Default: `orthogonal`.
            recurrent_regularizer: Regularizer function applied to the
                `recurrent_kernel` weights matrix. Default: `None`.
            recurrent_constraint: Constraint function applied to the `recurrent_kernel`
                weights matrix. Default: `None`.
            dropout: Float between 0 and 1. Fraction of the units to drop for the
                linear transformation of the inputs. Default: 0.
            recurrent_dropout: Float between 0 and 1. Fraction of the units to drop for
                the linear transformation of the recurrent state. Default: 0.
            reset_after: GRU convention (whether to apply reset gate after or
                before matrix multiplication). False = "before",
                True = "after" (default and cuDNN compatible).
            use_dropout (bool): Whether to use dropout on input features. Default is False.
            rate (float): Float between 0 and 1. Fraction of the input units to drop.
            noise_shape: 1D integer tensor representing the shape of the
                binary dropout mask that will be multiplied with the input.
            seed (int): A Python integer to use as random seed.
        """
        super(HamNetFingerprintGenerator, self).__init__(**kwargs)
        self.units = int(units)
        self.units_attend = int(units_attend)
        self.use_bias = bool(use_bias)
        self.use_dropout = use_dropout
        self.depth = int(depth)
        self.pooling_method = pooling_method
        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
        gru_args = {"recurrent_activation": recurrent_activation,
                    "use_bias": use_bias, "kernel_initializer": kernel_initializer,
                    "recurrent_initializer": recurrent_initializer, "bias_initializer": bias_initializer,
                    "kernel_regularizer": kernel_regularizer, "recurrent_regularizer": recurrent_regularizer,
                    "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                    "recurrent_constraint": recurrent_constraint, "bias_constraint": bias_constraint,
                    "dropout": dropout, "recurrent_dropout": recurrent_dropout, "reset_after": reset_after}
        self.pool_nodes = PoolingNodes(pooling_method=self.pooling_method)
        self.vertex2mol = DenseEmbedding(
            units=units, activation=activation, use_bias=use_bias, **kernel_args)

        self.readouts = [HamNetGlobalReadoutAttend(
            units=units_attend, activation=activation, activation_last="elu", use_bias=use_bias,
            use_dropout=use_dropout, rate=rate, noise_shape=noise_shape, seed=seed,
            **kernel_args) for _ in range(self.depth)]

        self.unions = [ks.layers.GRUCell(
            units=units, activation="tanh", **gru_args) for _ in range(self.depth)]
        self.final_activ = ActivationEmbedding(activation=activation,
                                               activity_regularizer=activity_regularizer)

    def build(self, input_shape):
        """Build layer."""
        super(HamNetFingerprintGenerator, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor): Node embeddings of shape `(batch, [N], F)`

        Returns:
            tf.RaggedTensor: Embedding tensor of pooled node attentions of shape `(batch, F)`
        """
        self.assert_ragged_input_rank(inputs)
        hv_ftr = inputs
        hm_ftr = self.vertex2mol(hv_ftr, **kwargs)
        hm_ftr = self.pool_nodes(hm_ftr, **kwargs)
        alignments = []
        for i in range(self.depth):
            mm_ftr, align = self.readouts[i]([hm_ftr, hv_ftr], **kwargs)
            # alignments.append(align)
            hm_ftr, _ = self.unions[i](mm_ftr, hm_ftr, **kwargs)
            hm_ftr = self.final_activ(hm_ftr, **kwargs)
        return hm_ftr

    def get_config(self):
        """Update layer config."""
        config = super(HamNetFingerprintGenerator, self).get_config()
        config.update({"use_bias": self.use_bias, "units": self.units, "units_attend": self.units_attend,
                       "use_dropout": self.use_dropout, "depth": self.depth, "pooling_method": self.pooling_method})
        conf_sub = self.vertex2mol.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
            config.update({x: conf_sub[x]})
        if len(self.unions) > 0:
            conf_gru = self.unions[0].get_config()
            for x in ["recurrent_activation", "recurrent_initializer", "recurrent_regularizer", "recurrent_constraint",
                      "dropout", "recurrent_dropout", "reset_after"]:
                config.update({x: conf_gru[x]})
        if len(self.readouts) > 0:
            conf_read = self.readouts[0].get_config()
            for x in ["use_dropout", "seed", "rate", "noise_shape", ]:
                config.update({x: conf_read[x]})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='HamNaiveDynMessage')
class HamNaiveDynMessage(GraphBaseLayer):
    r"""Message passing block from `HamNet <https://arxiv.org/abs/2105.03688>`_ which makes use of attention.
    The naming convention follows the authors `implementation <https://github.com/PKUterran/MoleculeClub>`_.
    The layer computes the following, let :math:`\mathbf{h}`, :math:`\mathbf{\epsilon}_{ij}` be node, edge features
    and :math:`\mathbf{q}`, :math:`\mathbf{p}` be (generalized) node coordinates and momentum. With
    :math:`\mathbf{p}_{ij} = \mathbf{p}_{j} - \mathbf{p}_{i}` and
    :math:`\mathbf{q}_{ij} = \mathbf{q}_{j} - \mathbf{q}_{i}` the attention coefficients read:

    .. math::
        \mathbf{a}_{ij} = \mathbf{w}^T \left(\mathbf{p}_{ij} \; || \; \mathbf{q}_{ij} \; ||
        \mathbf{\epsilon}_{ij} \right)

    and the new node update or message, using the attention coefficients
    :math:`\alpha_{ij} = \, \text{softmax}(\{\mathbf{a}_{ij} \; | \; j \in \mathcal{N}(i)\})`:

    .. math::
        \mathbf{m}_{v} = \sigma \; \sum_{j \in \mathcal{N}(i)} \; \alpha_{ij} \;  \sigma
        \left[ \; \mathbf{h}_j \; \mathbf{W}^T \;  \right]

    and edge updates:

    .. math::
        \mathbf{m}_{e} = \sigma \left[ \left(\; \mathbf{h}_i \; || \; \mathbf{q}_{ij} \; || \;
        \mathbf{h}_j \; \right) \mathbf{W}^T \right]

    the layer returns :math:`\mathbf{m}_{v}` and :math:`\mathbf{m}_{e}`.

    """
    def __init__(self,
                 units,
                 units_edge,
                 activation="kgcnn>leaky_relu",
                 activation_last="elu",
                 use_bias=True,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 use_dropout=False,
                 rate=None, noise_shape=None, seed=None,
                 **kwargs):
        r"""Initialize layer.

        Args:
            units (int): Units for the linear transformation of node features before attention.
            units_edge (int): Units for :obj:`Dense` layer for edge updates.
            activation (str, dict): Activation. Default is "kgcnn>leaky_relu".
            activation_last (str, dict): Last activation for messages. Default is "elu".
            use_bias (bool): Boolean, whether the layer uses a bias vector. Default is True.
            kernel_initializer: Initializer for the `kernel` weights matrix. Default is "glorot_uniform".
            bias_initializer: Initializer for the bias vector. Default is "zeros".
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix. Default is None.
            bias_regularizer: Regularizer function applied to the bias vector. Default is None.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation"). Default is None.
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix. Default is None.
            bias_constraint: Constraint function applied to the bias vector. Default is None.
            use_dropout (bool): Whether to use dropout on input features. Default is False.
            rate (float): Float between 0 and 1. Fraction of the input units to drop.
            noise_shape: 1D integer tensor representing the shape of the
                binary dropout mask that will be multiplied with the input.
            seed (int): A Python integer to use as random seed.
        """
        super(HamNaiveDynMessage, self).__init__(**kwargs)
        self.units = int(units)
        self.units_edge = int(units_edge)
        self.use_bias = use_bias
        self.use_dropout = use_dropout
        kernel_args = {"kernel_regularizer": kernel_regularizer,
                       "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
        self.gather_v = GatherEmbeddingSelection(selection_index= [0, 1])
        self.gather_p = GatherEmbeddingSelection(selection_index= [0, 1])
        self.gather_q = GatherEmbeddingSelection(selection_index= [0, 1])
        self.lazy_sub_p = LazySubtract()
        self.lazy_sub_q = LazySubtract()
        # self.lay_concat = LazyConcatenate(axis=-1)
        self.lay_concat_align = LazyConcatenate(axis=-1)
        self.lay_concat_edge = LazyConcatenate(axis=-1)
        if self.use_dropout:
            self.dropout_layer = DropoutEmbedding(rate=rate, noise_shape=noise_shape, seed=seed)
        self.dense_attend = DenseEmbedding(units=units, use_bias=use_bias, activation=activation, **kernel_args)
        self.dense_align = DenseEmbedding(1, activation="linear", use_bias=use_bias, **kernel_args)
        self.dense_e = DenseEmbedding(units=units_edge, activation=activation, use_bias=use_bias, **kernel_args)
        self.pool_attention = PoolingLocalEdgesAttention()
        self.final_activ = ActivationEmbedding(activation=activation_last,
                                               activity_regularizer=activity_regularizer)

    def build(self, input_shape):
        """Build layer."""
        super(HamNaiveDynMessage, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [hv_ftr, he_ftr, p_ftr, q_ftr, edge_index]

                - hv_ftr (tf.RaggedTensor): Node features of shape `(batch, [N], F)`
                - he_ftr (tf.RaggedTensor): Edge features of shape `(batch, [M], F)`
                - p_ftr (tf.RaggedTensor): Momentum node features of shape `(batch, [N], F)`
                - q_ftr (tf.RaggedTensor): Positional node features of shape `(batch, [N], F)`
                - edge_index (tf.RaggedTensor): Edge connection index list of shape `(batch, [M], 2)`

        Returns:
            list: [mv_ftr, me_ftr]

                - mv_ftr (tf.RaggedTensor): Node feature updates of shape `(batch, [N], F)`
                - me_ftr (tf.RaggedTensor): Edge feature updates of shape `(batch, [M], F)`
        """
        hv_ftr, he_ftr, p_ftr, q_ftr, edi = inputs
        if self.use_dropout:
            hv_ftr = self.dropout_layer(hv_ftr, **kwargs)
            he_ftr = self.dropout_layer(he_ftr, **kwargs)
        hv_u_ftr, hv_v_ftr = self.gather_v([hv_ftr, edi], **kwargs)
        q_u_ftr, q_v_ftr = self.gather_p([q_ftr, edi], **kwargs)
        p_u_ftr, p_v_ftr = self.gather_q([p_ftr, edi], **kwargs)
        p_uv_ftr = self.lazy_sub_p([p_v_ftr, p_u_ftr], **kwargs)
        q_uv_ftr = self.lazy_sub_p([q_v_ftr, q_u_ftr], **kwargs)

        attend_ftr = self.dense_attend(hv_v_ftr, **kwargs)

        align_ftr = self.lay_concat_align([p_uv_ftr, q_uv_ftr, he_ftr], **kwargs)
        align_ftr = self.dense_align(align_ftr, **kwargs)
        mv_ftr = self.pool_attention([hv_ftr, attend_ftr, align_ftr, edi], **kwargs)
        mv_ftr = self.final_activ(mv_ftr, **kwargs)

        me_ftr = self.lay_concat_edge([hv_u_ftr, p_uv_ftr, q_uv_ftr, hv_v_ftr], **kwargs)
        me_ftr = self.dense_e(me_ftr)

        return mv_ftr, me_ftr

    def get_config(self):
        """Update layer config."""
        config = super(HamNaiveDynMessage, self).get_config()
        config.update({"use_bias": self.use_bias, "units": self.units, "use_dropout": self.use_dropout})
        conf_sub = self.dense_attend.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
            config.update({x: conf_sub[x]})
        if self.use_dropout:
            conf_drop = self.dropout_layer.get_config()
            for x in ["rate", "noise_shape", "seed"]:
                config.update({x: conf_drop[x]})
        conf_last = self.final_activ.get_config()
        config.update({"activation_last": conf_last["activation"]})
        return config


# @ks.utils.register_keras_serializable(package='kgcnn', name='HamiltonEngine')
# class HamiltonEngine(GraphBaseLayer):
#     def __init__(self,
#                  units,
#                  activation="kgcnn>leaky_relu",
#                  use_bias=True,
#                  kernel_regularizer=None,
#                  bias_regularizer=None,
#                  activity_regularizer=None,
#                  kernel_constraint=None,
#                  bias_constraint=None,
#                  kernel_initializer='glorot_uniform',
#                  bias_initializer='zeros',
#                  **kwargs):
#         """Initialize layer."""
#         super(HamiltonEngine, self).__init__(**kwargs)
#         self.units = int(units)
#         self.use_bias = bool(use_bias)
#         kernel_args = {"kernel_regularizer": kernel_regularizer,
#                        "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
#                        "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
#                        "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
#         self.dense_atom = DenseEmbedding(units=units, activation="tanh", use_bias=use_bias, **kernel_args)
#         self.dense_edge = DenseEmbedding(units=units, activation="tanh", use_bias=use_bias, **kernel_args)
#
#     def build(self, input_shape):
#         """Build layer."""
#         super(HamiltonEngine, self).build(input_shape)
#
#     def call(self, inputs, **kwargs):
#         """Forward pass.
#
#         Args:
#             inputs: [nodes, edge, edge_indices]
#
#                 - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
#
#         Returns:
#             tf.RaggedTensor: Embedding tensor of pooled node attentions of shape (batch, F)
#         """
#         atom_ftr, bond_ftr, edge_idx = inputs
#         hv_ftr = self.dense_atom(atom_ftr, **kwargs)
#         he_ftr = self.dense_edge(bond_ftr, **kwargs)
#         return
#
#     def get_config(self):
#         """Update layer config."""
#         config = super(HamiltonEngine, self).get_config()
#         config.update({"use_bias": self.use_bias, "units": self.units,})
#         return config
